import pytest
from django.db import connection
from django.test.utils import CaptureQueriesContext
from django.urls import reverse

from emeis.core import models
from emeis.core.visibilities import filter_queryset_for


class TestACLVisibility:
    @filter_queryset_for(models.ACL)
    def filter_queryset_for_acl(self, queryset, request):
        return queryset.filter(role__slug="foo")


@pytest.mark.parametrize("include", ["acls", ""])
@pytest.mark.freeze_time("2017-05-21")
def test_invisible_acl_in_users(
    admin_client,
    set_visibilities,
    user_factory,
    role_factory,
    acl_factory,
    include,
):
    role_foo = role_factory(slug="foo")
    role_bar = role_factory(slug="bar")
    url = reverse("user-list")

    user1, user2 = user_factory.create_batch(2)
    # both users have both roles in an ACL. However only the "foo"
    # ACL should be returned in the includes and relationships
    acl_factory(role=role_foo, user=user1)
    acl_factory(role=role_foo, user=user2)
    acl_factory(role=role_bar, user=user1)
    acl_factory(role=role_bar, user=user2)

    set_visibilities(["emeis.core.tests.test_prefetch_visibility.TestACLVisibility"])

    with CaptureQueriesContext(connection) as context:
        resp = admin_client.get(url, {"include": include})

    assert len(context.captured_queries) == 2

    collected_acl_rels = [
        acl["id"]
        for u in resp.json()["data"]
        for acl in u["relationships"]["acls"]["data"]
    ]
    related_roles = models.Role.objects.filter(acls__in=collected_acl_rels)
    assert role_bar not in related_roles

    if include:
        collected_acl_includes = [
            inc["id"] for inc in resp.json()["included"] if inc["type"] == "acls"
        ]
        included_roles = models.Role.objects.filter(acls__in=collected_acl_includes)
        assert role_bar not in included_roles


@pytest.mark.parametrize("include", ["acls", "acls.role", "acls.role,acls.scope"])
@pytest.mark.freeze_time("2017-05-21")
def test_number_of_queries(
    admin_client,
    set_visibilities,
    user_factory,
    role_factory,
    acl_factory,
    include,
):
    role_foo = role_factory(slug="foo")
    role_bar = role_factory(slug="bar")
    url = reverse("user-list")

    def create_some():
        user1, user2 = user_factory.create_batch(2)
        # both users have both roles in an ACL. However only the "foo"
        # ACL should be returned in the includes and relationships
        acl_factory(role=role_foo, user=user1)
        acl_factory(role=role_foo, user=user2)
        acl_factory(role=role_bar, user=user1)
        acl_factory(role=role_bar, user=user2)
        return user1, user2

    set_visibilities(["emeis.core.tests.test_prefetch_visibility.TestACLVisibility"])

    create_some()

    with CaptureQueriesContext(connection) as context:
        admin_client.get(url, {"include": include})
    num_queries = len(context.captured_queries)

    create_some()

    with CaptureQueriesContext(connection) as context:
        admin_client.get(url, {"include": include})

    # We need to be sure that we don't create N+1 problems: Regardless of
    # includes, and regardless of number of users, the number of queries
    # must not change
    assert num_queries == len(context.captured_queries)



# Generated by Django 2.2.24 on 2021-07-26 09:06

import django.contrib.auth.models
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("emeis_core", "0004_use_abstract_base_user"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="user",
            managers=[
                ("objects", django.contrib.auth.models.UserManager()),
            ],
        ),
    ]



from unicodedata import normalize

import pytest
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.base_user import BaseUserManager
from django.utils import translation
from hypothesis import given
from hypothesis.strategies import emails, text
from mptt.exceptions import InvalidMove

from ..models import Scope, User


def test_user_model():
    user = User(
        username="foobar", email="foo@example.com", first_name="Foo", last_name="Bar"
    )
    assert user.get_full_name() == "Foo Bar"
    assert user.natural_key() == user.username
    assert user.get_email_field_name() == "email"
    assert user.is_anonymous is False
    assert user.is_authenticated is True


@given(username=text(), email=emails())
def test_user_model_normalization(username, email):
    username_normal = normalize("NFKC", username)
    email_normal = BaseUserManager.normalize_email(email)
    user = User(username=username, email=email, first_name="Foo", last_name="Bar")
    user.clean()
    assert user.username == username_normal
    assert user.email == email_normal


def test_scope_model(db):
    parent_scope = Scope.objects.create(name="parent scope")
    scope = Scope.objects.create(name="child scope", parent=parent_scope)

    assert scope.parent.name["en"] == "parent scope"

    parent_scope.parent = scope
    with pytest.raises(InvalidMove):
        parent_scope.save()


def test_scope_deletion(db, scope_factory):
    root = scope_factory()  # not deleted
    child = scope_factory(parent=root)
    grandchild = scope_factory(parent=child)
    scope_factory(parent=grandchild)
    scope_factory(parent=grandchild)

    other_child = scope_factory(parent=root)  # not deleted
    scope_factory(parent=other_child)  # not deleted

    child.delete()

    assert Scope.objects.count() == 3
    assert list(Scope.objects.root_nodes()) == [root]


def test_can_authenticate(db, user):
    # Test whether the authentication mechanism works correctly
    user.set_password("test_password")
    user.save()

    backend = ModelBackend()

    request = None
    auth = backend.authenticate(
        request, username=user.username, password="test_password"
    )
    assert auth == user


def test_scope_hierarchical_name(db, scope_factory):
    root = scope_factory()
    child = scope_factory(parent=root)
    grandchild = scope_factory(parent=child)

    assert (
        grandchild.full_name
        == f"{root.name} \u00bb {child.name} \u00bb {grandchild.name}"
    )


@pytest.mark.parametrize(
    "language, expected",
    [
        ("de", "DE ROOT » FR CHILD » DE GRANDCHILD"),
        ("fr", "DE ROOT » FR CHILD » FR GRANDCHILD"),
    ],
)
def test_scope_hierarchical_name_fallbacks(
    db, language, scope_factory, expected, settings
):
    settings.LANGUAGE_CODE = "de"
    settings.LANGUAGES = [("de", "de"), ("fr", "fr")]
    settings.LOCALIZED_FIELDS_FALLBACKS = {"fr": ["de"], "de": ["fr"]}
    with translation.override(language):
        root = scope_factory()
        root.name = {"de": "DE ROOT"}
        root.save()
        child = scope_factory(parent=root)
        child.name = {"fr": "FR CHILD"}
        child.save()
        grandchild = scope_factory(parent=child)
        grandchild.name = {"de": "DE GRANDCHILD", "fr": "FR GRANDCHILD"}
        grandchild.save()

        assert str(grandchild.full_name) == expected


@pytest.mark.parametrize("language", ["de", "fr"])
def test_scope_fullname_when_forced_language(db, language, scope_factory, settings):
    settings.LANGUAGE_CODE = "de"
    settings.LANGUAGES = [("de", "de"), ("fr", "fr")]
    settings.EMEIS_FORCE_MODEL_LOCALE = {"scope": "de"}

    with translation.override(language):
        root = scope_factory(name="DE ROOT")
        child = scope_factory(parent=root, name={"de": "DE CHILD", "fr": "FR CHILD"})
        grandchild = scope_factory(
            parent=child, name={"de": "DE GRANDCHILD", "fr": "FR GRANDCHILD"}
        )

        # Trigger pre_save `set_full_name()` hook
        grandchild.save()
        assert grandchild.full_name.de == "DE ROOT » DE CHILD » DE GRANDCHILD"
        assert grandchild.full_name.fr == ""


def test_update_full_name_of_child(db, scope_factory):
    root = scope_factory(parent=None, name="r")
    child = scope_factory(parent=root, name="c")
    grandchild = scope_factory(parent=child, name="g")

    # full name should be r -> c -> g
    assert str(grandchild.full_name) == "r » c » g"

    sibling = scope_factory(parent=root, name="s")

    # move parent of grandchild - this should trigger an
    # update on the grandchild's full name
    child.parent = sibling
    child.save()

    grandchild.refresh_from_db()
    assert str(grandchild.full_name) == "r » s » c » g"



import pytest
from django.urls import reverse
from generic_permissions.config import ObjectPermissionsConfig, PermissionsConfig
from generic_permissions.permissions import object_permission_for, permission_for
from rest_framework.status import (
    HTTP_200_OK,
    HTTP_201_CREATED,
    HTTP_204_NO_CONTENT,
    HTTP_403_FORBIDDEN,
)

from emeis.core.models import User

TIMESTAMP = "2017-05-21T11:25:41.123840Z"


@pytest.mark.freeze_time(TIMESTAMP)
@pytest.mark.parametrize(
    "method,status",
    [
        ("post", HTTP_201_CREATED),
        ("patch", HTTP_200_OK),
        ("delete", HTTP_204_NO_CONTENT),
    ],
)
@pytest.mark.parametrize("use_admin_client", [True, False])
def test_permission(
    user_factory,
    admin_user,
    admin_client,
    client,
    method,
    status,
    use_admin_client,
    reset_config_classes,
):
    client = admin_client if use_admin_client else client

    class CustomPermission:
        @permission_for(User)
        def has_permission_for_user(self, request):
            if request.user.username == "admin" or request.data["phone"] == "232355":
                return True
            return False

        @object_permission_for(User)
        def has_object_permission_for_user(self, request, instance):
            assert isinstance(instance, User)
            if request.user.username == "admin":
                return True
            return False

    PermissionsConfig.register_handler_class(CustomPermission)
    ObjectPermissionsConfig.register_handler_class(CustomPermission)

    user = user_factory(username="foo")

    url = reverse("user-list")

    data = {
        "data": {
            "type": "users",
            "attributes": {
                "meta": {},
                "username": "mark48",
                "first-name": "Amanda",
                "last-name": "Gallagher",
                "email": "banderson@example.com",
                "phone": None,
                "language": "en",
                "address": None,
                "city": None,
                "zip": None,
                "is-active": True,
            },
        }
    }

    if method in ["patch", "delete"]:
        url = reverse("user-detail", args=[user.pk])
        data["data"]["id"] = str(user.pk)

    if not use_admin_client and method == "patch":
        data["data"]["attributes"]["phone"] = "232355"

    response = getattr(client, method)(url, data=data)

    if not use_admin_client:
        assert response.status_code == HTTP_403_FORBIDDEN
        return

    assert response.status_code == status
    if method == "post":
        result = response.json()
        assert result["data"]["attributes"]["date-joined"] == TIMESTAMP
    elif method == "patch":
        user.refresh_from_db()
        assert user.username == "mark48"



import pytest
from django.urls import reverse
from django.utils import translation
from rest_framework import status

from emeis.core.models import User


@pytest.mark.parametrize(
    "user_attribute",
    [
        lambda u: u.first_name,
        lambda u: u.last_name,
        lambda u: u.acls.first().scope.name,
        lambda u: u.acls.first().role.name,
    ],
)
@pytest.mark.parametrize(
    "partial_search", [lambda val: val[:-2], lambda val: val[2:], lambda val: val]
)
def test_search_users(admin_client, acl_factory, user_attribute, partial_search):
    users_list = [acl.user for acl in acl_factory.create_batch(5)]

    resp = admin_client.get(
        reverse("user-list"),
        {"filter[search]": partial_search(str(user_attribute(users_list[0])))},
    )
    returned_user_ids = [us["id"] for us in resp.json()["data"]]

    assert str(users_list[0].pk) in returned_user_ids
    # ensure we don't just return the full user list
    assert len(returned_user_ids) < len(users_list)


@pytest.mark.parametrize(
    "filter_name, expect_result",
    [("hasRole", True), ("has_role", True), ("hasrole", False)],
)
def test_user_has_role(admin_client, acl_factory, filter_name, expect_result):
    users_list = [acl.user for acl in acl_factory.create_batch(3)]

    resp = admin_client.get(
        reverse("user-list"),
        {f"filter[{filter_name}]": users_list[2].acls.first().role_id},
    )

    if expect_result:
        ret_users = [us["attributes"]["username"] for us in resp.json()["data"]]
        expected = [str(users_list[2].username)]

        assert expected == ret_users
    else:
        assert resp.status_code == status.HTTP_400_BAD_REQUEST


@pytest.mark.parametrize(
    "filter_value, expect_count",
    [("true", 4), ("false", 2)],
)
def test_user_is_active_filter(admin_client, user_factory, filter_value, expect_count):
    user_factory.create_batch(3, is_active=True)
    user_factory.create_batch(2, is_active=False)

    resp = admin_client.get(
        reverse("user-list"),
        {"filter[is_active]": filter_value},
    )

    assert len(resp.json()["data"]) == expect_count


@pytest.mark.parametrize(
    "filter_field, model_attr, expect_result",
    [
        ("id__in", "pk", True),
        ("id", "pk", True),
        ("username", "username", True),
        ("username", "email", False),
        ("email__in", "email", True),
        ("email__in", "username", False),
    ],
)
def test_declared_filters(
    admin_client, user_factory, filter_field, model_attr, expect_result
):
    user1, _ = user_factory.create_batch(2)

    attr_value = getattr(user1, model_attr)

    resp = admin_client.get(
        reverse("user-list"), {f"filter[{filter_field}]": attr_value}
    )

    ret_users = [us["attributes"]["username"] for us in resp.json()["data"]]

    if expect_result:
        assert [user1.username] == ret_users
    else:
        assert ret_users == []


@pytest.mark.parametrize(
    "filter_field, model_attr",
    [
        ("user__in", "user_id"),
        ("role__in", "role_id"),
        ("scope__in", "scope_id"),
    ],
)
def test_acl_filters(admin_client, acl_factory, filter_field, model_attr):
    acl1, acl2, _ = acl_factory.create_batch(3)

    attr_value_1 = getattr(acl1, model_attr)
    attr_value_2 = getattr(acl2, model_attr)

    resp = admin_client.get(
        reverse("acl-list"),
        {f"filter[{filter_field}]": f"{attr_value_1},{attr_value_2}"},
    )

    ret_acls = [acl["id"] for acl in resp.json()["data"]]
    assert set([str(acl1.pk), str(acl2.pk)]) == set(ret_acls)


def test_scope_id_filter(admin_client, scope_factory):
    scope1, _, scope3 = scope_factory.create_batch(3)

    resp = admin_client.get(
        reverse("scope-list"),
        {"filter[id__in]": ",".join([str(scope1.pk), str(scope3.pk)])},
    )

    ret_scopes = [us["id"] for us in resp.json()["data"]]

    assert set(ret_scopes) == set([str(scope1.pk), str(scope3.pk)])


@pytest.mark.parametrize("sort", ["email", "-email"])
def test_user_ordering_case_insensitive(admin_client, admin_user, user_factory, sort):
    emails = [
        "Aaaa@example.com",
        "Zzzzz@example.com",
        "aaaaa@example.com",
        "m@example.com",
    ]
    for email in emails:
        user_factory.create(email=email)

    resp = admin_client.get(reverse("user-list"), {"sort": sort})

    expect_emails = sorted(
        emails + [admin_user.user.email],
        key=lambda e: e.lower(),
        reverse=sort.startswith("-"),
    )

    assert expect_emails == [d["attributes"]["email"] for d in resp.json()["data"]]


@pytest.mark.parametrize("sort", ["metainfo__position", "-metainfo__position"])
def test_ordering_metainfo(admin_client, admin_user, user_factory, sort, settings):
    settings.EMEIS_META_FIELDS = ["position"]
    admin_user.user.metainfo = {"position": "d"}
    admin_user.user.save()
    positions = ["B", "a", "C"]
    for position in positions:
        user_factory.create(metainfo={"position": position})

    resp = admin_client.get(reverse("user-list"), {"sort": sort})

    expect = sorted(
        positions + ["d"], key=lambda e: e.lower(), reverse=sort.startswith("-")
    )

    assert expect == [
        d["attributes"]["metainfo"].get("position") for d in resp.json()["data"]
    ]


@pytest.mark.parametrize("sort", ["username", "-username"])
def test_user_ordering_case_sensitive(admin_client, admin_user, user_factory, sort):
    user_factory.create_batch(5)

    resp = admin_client.get(reverse("user-list"), {"sort": sort})

    expected = list(
        User.objects.all().order_by(sort).values_list("username", flat=True)
    )

    assert expected == [d["attributes"]["username"] for d in resp.json()["data"]]


@pytest.mark.parametrize(
    "force_lang, search_term, expect_result",
    [
        ("de", "deutscher", 1),
        ("de", "english", 0),
        ("en", "deutscher", 0),
        ("en", "english", 1),
        (None, "deutscher", 0),
        (None, "english", 1),
    ],
)
def test_search_monolingual(
    settings, admin_client, role_factory, force_lang, search_term, expect_result
):
    role_factory(name={"de": "deutscher name", "en": "english name"})

    if force_lang:
        settings.EMEIS_FORCE_MODEL_LOCALE = {"role": force_lang}

    with translation.override("en"):
        resp = admin_client.get(reverse("role-list"), {"filter[search]": search_term})

    assert len(resp.json()["data"]) == expect_result


@pytest.mark.parametrize(
    "search_term, expect_result",
    [
        ("foobar", 1),
        ("bar", 2),
        ("buzzybuzz", 0),
    ],
)
def test_search_metainfo(
    settings, admin_client, user_factory, search_term, expect_result
):
    settings.EMEIS_META_FIELDS = ["position"]
    user_factory(metainfo={"position": "bar"})
    user_factory(metainfo={"position": "baz"})
    user_factory(metainfo={"blubb": "bar"})
    user_factory(metainfo={"position": "Foobar"})

    resp = admin_client.get(reverse("user-list"), {"filter[search]": search_term})

    assert len(resp.json()["data"]) == expect_result



# Generated by Django 2.2.26 on 2022-01-31 12:19

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("emeis_core", "0005_manager_on_user_model"),
    ]

    operations = [
        migrations.RenameField(model_name="acl", old_name="meta", new_name="metainfo"),
        migrations.RenameField(
            model_name="permission", old_name="meta", new_name="metainfo"
        ),
        migrations.RenameField(model_name="role", old_name="meta", new_name="metainfo"),
        migrations.RenameField(
            model_name="scope", old_name="meta", new_name="metainfo"
        ),
        migrations.RenameField(model_name="user", old_name="meta", new_name="metainfo"),
        migrations.AlterField(
            model_name="acl",
            name="metainfo",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=dict, verbose_name="metainfo"
            ),
        ),
        migrations.AlterField(
            model_name="permission",
            name="metainfo",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=dict, verbose_name="metainfo"
            ),
        ),
        migrations.AlterField(
            model_name="role",
            name="metainfo",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=dict, verbose_name="metainfo"
            ),
        ),
        migrations.AlterField(
            model_name="scope",
            name="metainfo",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=dict, verbose_name="metainfo"
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="metainfo",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=dict, verbose_name="metainfo"
            ),
        ),
    ]



# Generated by Django 3.2.15 on 2022-09-23 09:13

from django.db import migrations
import localized_fields.fields.char_field

from emeis.core import models


def set_full_name(apps, schema_editor):
    scope_model = apps.get_model("emeis_core.scope")
    for scope in scope_model.objects.all().iterator():
        # explicitly trigger the set_full_name signal handler
        models.set_full_name(instance=scope, sender=set_full_name)
        scope.save()


class Migration(migrations.Migration):
    dependencies = [
        ("emeis_core", "0009_alter_scope_parent"),
    ]

    operations = [
        migrations.AddField(
            model_name="scope",
            name="full_name",
            field=localized_fields.fields.char_field.LocalizedCharField(
                blank=True, null=True, required=[], verbose_name="scope name"
            ),
        ),
        migrations.RunPython(set_full_name, migrations.RunPython.noop),
    ]



from django.conf import settings
from django.contrib.postgres.fields.hstore import KeyTransform
from django.core.exceptions import FieldDoesNotExist
from django.db.models import TextField
from django.db.models.functions import Cast, Lower
from django_filters import FilterSet
from django_filters.filters import CharFilter
from localized_fields.fields import LocalizedField
from rest_framework import filters

from emeis.core import utils
from emeis.core.models import ACL, Scope, User


class EmeisSearchFilter(filters.SearchFilter):
    """Custom search filter for Emeis.

    This adds two things:
    - "monolingual" search support via EMEIS_FORCE_MODEL_LOCALE
    - search support for EMEIS_META_FIELDS
    """

    multilingual_search_fields = []

    def get_search_fields(self, view, request):
        """Return search fields for the current view.

        The search fields are defined on the viewset as normal, but if the
        model is configured as a "forced monolingual" model, we use the default
        language instead for searching.

        """

        model_name = view.queryset.model.__name__.lower()

        meta_fields = [f"metainfo__{f}" for f in settings.EMEIS_META_FIELDS]

        if model_name not in settings.EMEIS_FORCE_MODEL_LOCALE:
            default_fields = super().get_search_fields(view, request)
            return list(default_fields) + meta_fields if default_fields else meta_fields

        forced_lang = settings.EMEIS_FORCE_MODEL_LOCALE[model_name]

        return [
            f"{field}__{forced_lang}"
            if field in view.multilingual_search_fields
            else field
            for field in view.search_fields
        ] + meta_fields


class UserFilterset(FilterSet):
    has_role = CharFilter(
        field_name="acls__role__pk",
        distinct=True,
    )

    class Meta:
        model = User

        fields = {
            "id": ["exact", "in"],
            "username": ["exact", "in"],
            "first_name": ["exact", "icontains", "contains"],
            "last_name": ["exact", "icontains", "contains"],
            "email": ["exact", "in"],
            "is_active": ["exact"],
        }


class ACLFilterset(FilterSet):
    class Meta:
        model = ACL
        fields = {
            "id": ["exact", "in"],
            "user": ["exact", "in"],
            "scope": ["exact", "in"],
            "role": ["exact", "in"],
        }


class EmeisOrderingFilter(filters.OrderingFilter):
    """Custom ordering filter for Emeis.

    This adds two things:
    - case insensitive ordering via "case_insensitive_ordering_fields"
    - ordering by metainfo fields
    """

    def get_valid_fields(self, queryset, view, context=None):
        ordering_fields = super().get_valid_fields(queryset, view, context or {})
        return ordering_fields + [
            (f"metainfo__{field}", f"{field} in metainfo")
            for field in settings.EMEIS_META_FIELDS
        ]

    def _make_ordering_field(self, field, view):
        model_name = view.queryset.model.__name__
        lang = utils.forced_or_current_lang(model_name)

        desc = field.startswith("-")
        field_name = field[1:] if desc else field

        field_col = field_name
        try:
            model_field = view.queryset.model._meta.get_field(field_name)
            if isinstance(model_field, LocalizedField):
                field_col = KeyTransform(lang, field_name)
        except FieldDoesNotExist:
            # This happens with metainfo__foobar style lookups,
            # which we don't need to worry about here
            pass

        if field_name in self._get_case_insensitive_fields(view):
            field_expr = Lower(Cast(field_col, output_field=TextField()))
            return field_expr.desc() if desc else field_expr
        return field

    def get_ordering(self, request, queryset, view):
        ordering = super().get_ordering(request, queryset, view)
        if not ordering:
            return ordering
        return [self._make_ordering_field(field, view) for field in ordering]

    def _get_case_insensitive_fields(self, view):
        return getattr(view, "case_insensitive_ordering_fields", []) + [
            f"metainfo__{field}" for field in settings.EMEIS_META_FIELDS
        ]


class ScopeFilterset(FilterSet):
    class Meta:
        model = Scope
        fields = {
            "id": ["exact", "in"],
        }



from uuid import uuid4

import pyexcel
import pytest
from django.urls import reverse
from rest_framework.status import (
    HTTP_200_OK,
    HTTP_201_CREATED,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_405_METHOD_NOT_ALLOWED,
)


def test_me_200(db, acl, client):
    client.force_authenticate(user=acl.user)
    url = reverse("me-detail")

    response = client.get(url, data={"include": "acls"})

    assert response.status_code == HTTP_200_OK
    result = response.json()
    assert len(result["included"]) == 1
    assert (
        result["data"]["id"]
        == result["included"][0]["relationships"]["user"]["data"]["id"]
        == str(acl.user.pk)
    )
    assert result["included"][0]["id"] == str(acl.pk)


@pytest.mark.parametrize(
    "method,status_code",
    [
        ("post", HTTP_405_METHOD_NOT_ALLOWED),
        ("patch", HTTP_405_METHOD_NOT_ALLOWED),
        ("put", HTTP_405_METHOD_NOT_ALLOWED),
        ("head", HTTP_200_OK),
        ("options", HTTP_200_OK),
    ],
)
def test_me_405(db, acl, client, method, status_code):
    client.force_authenticate(user=acl.user)
    url = reverse("me-detail")

    response = getattr(client, method)(url)

    assert response.status_code == status_code


@pytest.mark.parametrize("list", [True, False])
def test_myacls_200(db, acl_factory, client, list):
    acl = acl_factory(role__slug="my-role")
    acl_factory()
    client.force_authenticate(user=acl.user)
    url = reverse("myacls-list")
    if not list:
        url = reverse("myacls-detail", args=[acl.pk])

    response = client.get(url, data={"include": "role"})

    assert response.status_code == HTTP_200_OK
    result = response.json()
    assert len(result["included"]) == 1
    if list:
        assert result["data"][0]["id"] == str(acl.pk)
        assert len(result["data"]) == 1
    else:
        assert result["data"]["id"] == str(acl.pk)


def test_myacls_404(db, user, client):
    client.force_authenticate(user=user)
    url = reverse("myacls-detail", args=[str(uuid4())])

    response = client.get(url)

    assert response.status_code == HTTP_404_NOT_FOUND


def test_myacls_403(db, acl_factory, user, client):
    acl = acl_factory()
    client.force_authenticate(user=user)
    url = reverse("myacls-detail", args=[str(acl.pk)])

    response = client.get(url)

    assert response.status_code == HTTP_403_FORBIDDEN


def test_user_search_filter(db, user_factory, client):
    user_factory(username="user1", city="Lucerne", zip="6000", last_name="Kafka")
    user_factory(username="user2", city="Berne", zip="5000", last_name="Smith")
    user_factory(username="user3", city="Zurich", zip="8000", last_name="Smith")

    url = reverse("user-list")

    response = client.get(url, {"filter[search]": "bern"})
    assert response.status_code == HTTP_200_OK
    result = response.json()
    assert len(result["data"]) == 1
    assert result["data"][0]["attributes"]["username"] == "user2"

    response = client.get(url, {"filter[search]": "8000"})
    assert response.status_code == HTTP_200_OK
    result = response.json()
    assert len(result["data"]) == 1
    assert result["data"][0]["attributes"]["username"] == "user3"

    response = client.get(url, {"filter[search]": "smith"})
    assert response.status_code == HTTP_200_OK
    result = response.json()
    assert len(result["data"]) == 2
    assert "user1" not in (
        result["data"][0]["attributes"]["username"],
        result["data"][1]["attributes"]["username"],
    )


def test_scope_search_filter(db, scope_factory, client):
    scope_factory(name="scope1")
    scope_factory(name={"de": "skop2", "en": "scope2"})
    scope_factory(name="scope3")

    url = reverse("scope-list")

    response = client.get(url, {"filter[search]": "skop"}, HTTP_ACCEPT_LANGUAGE="de")

    assert response.status_code == HTTP_200_OK
    result = response.json()
    assert len(result["data"]) == 1
    assert result["data"][0]["attributes"]["name"] == {
        "de": "skop2",
        "en": "scope2",
        "fr": "",
    }
    assert result["data"][0]["attributes"]["level"] == 0

    response = client.get(url, {"filter[search]": "scope"})
    assert response.status_code == HTTP_200_OK
    result = response.json()
    assert len(result["data"]) == 3


def test_scope_full_name_api(db, scope_factory, client):
    parent = scope_factory()
    child = scope_factory(parent=parent)

    url = reverse("scope-detail", args=[child.pk])
    response = client.get(url)

    full_name = response.json()["data"]["attributes"]["full-name"]
    assert full_name["en"] == child.full_name
    assert full_name["en"].startswith(parent.name["en"])
    assert full_name["en"].endswith(child.name["en"])


def test_cannot_write_level(db, client, settings, user):
    data = {
        "data": {
            "type": "scopes",
            "attributes": {"name": {"en": "test"}, "level": 400},
        }
    }

    resp = client.post(reverse("scope-list"), data=data)

    assert resp.status_code == HTTP_201_CREATED
    result = resp.json()
    assert result["data"]["attributes"]["level"] == 0


def test_acl_user_filter(db, user, acl_factory, client):
    acl = acl_factory(user=user)
    acl_factory()

    url = reverse("acl-list")

    response = client.get(url, {"filter[user]": acl.user.pk})
    assert response.status_code == HTTP_200_OK
    result = response.json()
    assert len(result["data"]) == 1
    assert result["data"][0]["id"] == str(acl.pk)
    assert result["data"][0]["relationships"]["user"]["data"]["id"] == str(acl.user.pk)


def test_acl_search_filter(db, acl_factory, client):
    acl = acl_factory(role__name="scope1", scope__name="scope1", user__username="user1")
    acl_factory(role__name="scope1", scope__name="scope2", user=acl.user)
    acl_factory(role__name="scope1", scope=acl.scope, user__username="user2")

    url = reverse("acl-list")

    response = client.get(url, {"filter[search]": "user1"})
    assert response.status_code == HTTP_200_OK
    result = response.json()
    assert len(result["data"]) == 2
    assert (
        result["data"][0]["relationships"]["user"]["data"]["id"]
        == result["data"][1]["relationships"]["user"]["data"]["id"]
        == str(acl.user.pk)
    )


def test_permission_role_filter(db, user, role_factory, permission_factory, client):
    permission = permission_factory()
    role = role_factory()
    role.permissions.add(permission)

    dummy_permission = permission_factory()
    dummy_role = role_factory()
    dummy_role.permissions.add(dummy_permission)

    url = reverse("permission-list")

    response = client.get(url, {"filter[roles]": role.pk})
    assert response.status_code == HTTP_200_OK
    result = response.json()
    assert len(result["data"]) == 1
    assert result["data"][0]["id"] == str(permission.pk)


@pytest.mark.parametrize("allow_anon", [True, False])
@pytest.mark.parametrize("method", ["post", "patch"])
def test_anonymous_writing(db, client, settings, user, allow_anon, method):
    settings.ALLOW_ANONYMOUS_WRITE = allow_anon
    if not allow_anon:
        settings.REST_FRAMEWORK["DEFAULT_PERMISSION_CLASSES"] = [
            "rest_framework.permissions.IsAuthenticatedOrReadOnly",
        ]

    data = {
        "data": {
            "type": "users",
            "attributes": {"username": "winstonsmith", "language": "en"},
        }
    }

    url = reverse("user-list")

    if method == "patch":
        data["data"]["id"] = str(user.pk)
        url = reverse("user-detail", args=[user.pk])

    resp = getattr(client, method)(url, data=data)
    assert (
        resp.status_code == HTTP_201_CREATED or HTTP_200_OK
        if allow_anon
        else HTTP_403_FORBIDDEN
    )


def test_user_export(client, user_factory, acl_factory, snapshot):
    user = user_factory()
    acl_factory.create_batch(9)
    acl_factory.create_batch(2, user=user)
    # this user should not be exported
    user_factory(is_active=False)

    url = reverse("user-export")

    response = client.get(url, {"filter[isActive]": "true"})

    assert response.status_code == HTTP_200_OK

    book = pyexcel.get_book(file_content=response.getvalue(), file_type="xlsx")
    sheet = book.bookdict.popitem()[1]

    snapshot.assert_match(sheet)


def test_order_scopes_by_full_name(client, scope_factory, settings):
    settings.LANGUAGES = [
        ("de", "german"),
        ("en", "english"),
    ]
    r1 = scope_factory(name={"de": "erster", "en": "secondary"})
    r2 = scope_factory(name={"de": "zweiter", "en": "primary"})

    r1child1 = scope_factory(name={"de": "alfa", "en": "mike"}, parent=r1)
    r1child2 = scope_factory(name={"de": "kilo", "en": "echo"}, parent=r1)
    r1child3 = scope_factory(name={"de": "zulu", "en": "bravo"}, parent=r1)

    r2child1 = scope_factory(name={"de": "delta", "en": "hotel"}, parent=r2)
    r2child2 = scope_factory(name={"de": "golf", "en": "foxtrot"}, parent=r2)

    expected_sorting = {
        "de": [r1, r1child1, r1child2, r1child3, r2, r2child1, r2child2],
        "en": [r2, r2child2, r2child1, r1, r1child3, r1child2, r1child1],
    }

    for lang in ["de", "en"]:
        resp = client.get(
            reverse("scope-list"),
            {"sort": "full_name"},
            HTTP_ACCEPT_LANGUAGE=lang,
        )
        received_names = [
            entry["attributes"]["full-name"][lang] for entry in resp.json()["data"]
        ]
        received_ids = [entry["id"] for entry in resp.json()["data"]]
        expected_ids = [str(entry.pk) for entry in expected_sorting[lang]]

        assert sorted(received_names) == received_names
        assert received_ids == expected_ids


@pytest.mark.parametrize("language", ["de", "fr"])
def test_sorted_scopes_when_forced_language(
    db, client, language, scope_factory, settings
):
    # If forced locale is used, the request language should play no
    # role in the resulting order
    settings.LANGUAGE_CODE = "de"
    settings.LANGUAGES = [("de", "de"), ("fr", "fr")]
    settings.EMEIS_FORCE_MODEL_LOCALE = {"scope": "de"}

    # contrary sorting: de: eins -> zwei, fr: deux -> un
    scope_factory(name={"de": "eins", "fr": "un"})
    scope_factory(name={"de": "zwei", "fr": "deux"})

    resp = client.get(
        reverse("scope-list"),
        {"sort": "full_name"},
        HTTP_ACCEPT_LANGUAGE=language,
    )

    received_names = [
        entry["attributes"]["full-name"]["de"] for entry in resp.json()["data"]
    ]

    assert received_names == ["eins", "zwei"]



import unicodedata
import uuid

from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser, UserManager
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.db import models
from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.utils import timezone, translation
from django.utils.translation import gettext_lazy as _
from localized_fields.fields import LocalizedCharField, LocalizedTextField
from mptt.models import MPTTModel, TreeForeignKey


def make_uuid():
    """Return a new random UUID value.

    This indirection is done for testing purposes, so test code can mock
    uuid.uuid4(). If we wouldn't do this, then the models would have a direct
    reference that doesn't get mocked away.

    We can't replace it with a lambda because Django Migrations can't handle them.
    """
    return uuid.uuid4()


def get_language_code():
    return settings.LANGUAGE_CODE


class BaseModel(models.Model):
    created_at = models.DateTimeField(auto_now_add=True, db_index=True)
    modified_at = models.DateTimeField(auto_now=True, db_index=True)
    created_by_user = models.ForeignKey("User", null=True, on_delete=models.SET_NULL)
    metainfo = models.JSONField(_("metainfo"), default=dict)

    class Meta:
        abstract = True


class UUIDModel(BaseModel):
    """
    Models which use uuid as primary key.

    Defined as emeis default
    """

    id = models.UUIDField(primary_key=True, default=make_uuid, editable=False)

    class Meta:
        abstract = True


class SlugModel(BaseModel):
    """
    Models which use a slug as primary key.

    Defined as Emeis default for configuration so it is possible
    to merge between developer and user configuration.
    """

    slug = models.SlugField(max_length=255, primary_key=True)

    class Meta:
        abstract = True


class User(UUIDModel, AbstractBaseUser):
    username_validator = UnicodeUsernameValidator()

    username = models.CharField(
        _("username"),
        max_length=255,
        unique=True,
        help_text=_(
            "Required. 255 characters or fewer. Letters, digits and @/./+/-/_ only."
        ),
        validators=[username_validator],
        error_messages={"unique": _("A user with that username already exists.")},
    )
    first_name = models.CharField(
        _("first name"), max_length=255, blank=True, null=True
    )
    last_name = models.CharField(_("last name"), max_length=255, blank=True, null=True)
    email = models.EmailField(_("email address"), blank=True, null=True)
    phone = models.CharField(_("phone"), max_length=100, blank=True, null=True)
    language = models.CharField(_("language"), max_length=2, default=get_language_code)
    address = models.CharField(_("address"), max_length=255, blank=True, null=True)
    city = LocalizedCharField(_("city"), max_length=255, blank=True, null=True)
    zip = models.CharField(_("zip"), max_length=10, blank=True, null=True)
    is_active = models.BooleanField(
        _("active"),
        default=True,
        help_text=_("Designates whether this user should be treated as active."),
    )
    date_joined = models.DateTimeField(_("date joined"), default=timezone.now)

    EMAIL_FIELD = "email"
    USERNAME_FIELD = "username"
    REQUIRED_FIELDS = ["email"]

    objects = UserManager()

    def __str__(self):
        return "%s (username=%s)" % (type(self).__name__, self.username)

    class Meta:
        verbose_name = _("user")
        verbose_name_plural = _("users")

    def clean(self):
        super().clean()
        setattr(self, self.USERNAME_FIELD, self.normalize_username(self.get_username()))

        self.email = UserManager.normalize_email(self.email)

    def get_full_name(self):
        """Return the first_name plus the last_name, with a space in between."""
        full_name = "%s %s" % (self.first_name, self.last_name)
        return full_name.strip()

    def get_username(self):
        """Return the username for this User."""
        return getattr(self, self.USERNAME_FIELD)

    def natural_key(self):
        return self.get_username()

    @classmethod
    def get_email_field_name(cls):
        return "email"

    @classmethod
    def normalize_username(cls, username):
        return (
            unicodedata.normalize("NFKC", username)
            if isinstance(username, str)
            else username
        )

    @property
    def is_anonymous(self):
        """
        Return False.

        This is a way of comparing User objects to anonymous users.
        """
        return False

    @property
    def is_authenticated(self):
        """
        Return True.

        This is a way to tell if the user has been authenticated in templates.
        """
        return True


class Scope(MPTTModel, UUIDModel):
    name = LocalizedCharField(_("scope name"), blank=False, null=False, required=False)

    full_name = LocalizedCharField(
        _("scope name"), blank=True, null=True, required=False
    )

    description = LocalizedTextField(
        _("scope description"), null=True, blank=True, required=False
    )
    parent = TreeForeignKey(
        "self",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        related_name="children",
    )
    is_active = models.BooleanField(default=True)

    def __str__(self):
        return f"{type(self).__name__} ({self.full_name}, pk={self.pk})"

    class Meta:
        ordering = ["full_name"]


@receiver(pre_save, sender=Scope)
def set_full_name(instance, sender, **kwargs):
    if kwargs.get("raw"):  # pragma: no cover
        # Raw is set while loading fixtures. In those
        # cases we don't want to mess with data that
        # may not be there yet
        return
    sep = "\u00bb"

    languages = [lang for lang, _ in settings.LANGUAGES]

    forced_lang = settings.EMEIS_FORCE_MODEL_LOCALE.get("scope", None)

    if forced_lang:
        # If scope is forced monolingual, do not fill non-forced language fields
        languages = [forced_lang]

    for lang in languages:
        with translation.override(lang):
            instance.full_name[lang] = str(instance.name)

    parent = instance.parent
    while parent:
        for lang in languages:
            with translation.override(lang):
                new_fullname = f"{parent.name} {sep} {instance.full_name[lang]}"
                instance.full_name[lang] = new_fullname
        parent = parent.parent

    if forced_lang:
        # Ensure only the "forced" language full_name is set, and nothing else
        full_name = instance.full_name[forced_lang]
        instance.full_name.clear()
        instance.full_name[forced_lang] = full_name

    # Force update of all children (recursively)
    for child in instance.children.all():
        # save() triggers the set_full_name signal handler, which will
        # recurse all the way down, updating the full_name
        child.save()


class Role(SlugModel):
    name = LocalizedCharField(_("role name"), blank=False, null=False, required=False)
    description = LocalizedTextField(
        _("role description"), null=True, blank=True, required=False
    )
    permissions = models.ManyToManyField("Permission", related_name="roles")

    def __str__(self):
        return f"{type(self).__name__} ({self.pk})"

    class Meta:
        ordering = ["slug"]


class Permission(SlugModel):
    name = LocalizedCharField(
        _("permission name"), blank=False, null=False, required=False
    )
    description = LocalizedTextField(
        _("permission description"), null=True, blank=True, required=False
    )


class ACL(UUIDModel):
    user = models.ForeignKey("User", on_delete=models.CASCADE, related_name="acls")
    scope = models.ForeignKey("Scope", on_delete=models.CASCADE, related_name="acls")
    role = models.ForeignKey("Role", on_delete=models.CASCADE, related_name="acls")

    def __str__(self):
        return f"{type(self).__name__} (username={self.user.username}, scope=self.scope, role={self.role.pk})"

    class Meta:
        unique_together = ["user", "scope", "role"]



import io
from tempfile import NamedTemporaryFile

import localized_fields.fields
import openpyxl
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.db.models import Prefetch
from django.http import FileResponse
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from generic_permissions.permissions import PermissionViewMixin
from generic_permissions.visibilities import VisibilityViewMixin
from rest_framework.decorators import action
from rest_framework.mixins import ListModelMixin, RetrieveModelMixin
from rest_framework.permissions import IsAuthenticated
from rest_framework.viewsets import GenericViewSet
from rest_framework_json_api import views

from emeis.core import utils

from . import filters, models, serializers


class BaseViewset(VisibilityViewMixin, PermissionViewMixin, views.ModelViewSet):
    def replace_prefetch(self, qs, lookup, viewset_to_use):
        """Replace prefetch with a new prefetch respecting visibilities.

        This also works if DRF didn't already add a prefetch itself.
        Pass in the lookup (name of the field to the related object)
        as well as the viewset from where the queryset should be pulled:
        When retrieving the queryset from a viewset, the visibilities
        are automatically applied.

        Return a new queryset with the prefetch replaced or added
        """

        # Hacky hacky - remove the conflicting prefetches that DRF added,
        # as otherwise the queries won't work anymore.
        # There's no "official" way to do this in the ORM, and overriding
        # DRF's get_queryset() method would be even more hacky.
        qs._prefetch_related_lookups = tuple(
            [p for p in qs._prefetch_related_lookups if lookup not in p]
        )

        # Now: prefetch using the visibility-filtered relation
        related_queryset = viewset_to_use(request=self.request).get_queryset()
        return qs.prefetch_related(Prefetch(lookup, queryset=related_queryset))


class MeViewSet(RetrieveModelMixin, GenericViewSet):
    """Me view returns current user."""

    serializer_class = serializers.MeSerializer
    permission_classes = [IsAuthenticated]

    def get_object(self, *args, **kwargs):
        return self.request.user


class MyACLViewSet(RetrieveModelMixin, ListModelMixin, GenericViewSet):
    """MyACL view returns current users ACLs."""

    queryset = models.ACL.objects.all()
    serializer_class = serializers.MyACLSerializer
    permission_classes = [IsAuthenticated]

    def get_object(self):
        pk = self.kwargs.pop("pk")
        acl = get_object_or_404(models.ACL, pk=pk)
        if acl.user != self.request.user:
            raise PermissionDenied()
        return acl

    def get_queryset(self):
        return self.queryset.filter(user=self.request.user)


class UserViewSet(BaseViewset):
    serializer_class = serializers.UserSerializer
    queryset = models.User.objects.all()

    filterset_class = filters.UserFilterset

    search_fields = [
        "username",
        "first_name",
        "last_name",
        "acls__role__name",
        "acls__scope__name",
        "email",
        "=zip",
        "city",
    ]

    multilingual_search_fields = [
        "city",
        "acls__role__name",
        "acls__scope__name",
    ]

    case_insensitive_ordering_fields = [
        "first_name",
        "last_name",
        "email",
        "address",
        "city",
    ]

    def get_queryset(self):
        qs = super().get_queryset()
        return self.replace_prefetch(qs, "acls", ACLViewSet)

    @action(methods=["get"], detail=False)
    def export(self, request):
        """Export user list as excel table."""
        queryset = self.filter_queryset(self.get_queryset())

        # we use a template to set the headers to bold and align the columns to fit on a DIN-A4 page
        workbook = openpyxl.load_workbook(settings.USER_EXPORT_TEMPLATE_PATH)
        sheet = workbook.active

        # set column headers
        sheet["A1"] = _("Name")
        sheet["B1"] = _("First Name")
        sheet["C1"] = _("Email")
        sheet["D1"] = _("Roles and scopes")

        for row, user in enumerate(queryset.iterator(), start=2):
            acl_string = "\n".join(
                [
                    f"{acl.role.name}: {acl.scope.full_name}"
                    for acl in user.acls.all()
                    .select_related("scope", "role")
                    .order_by("role__name")
                ]
            )

            sheet[f"A{row}"] = user.last_name
            sheet[f"B{row}"] = user.first_name
            sheet[f"C{row}"] = user.email
            sheet[f"D{row}"] = acl_string
            sheet[f"D{row}"].alignment = openpyxl.styles.Alignment(wrap_text=True)

        # save workbook as temporary file and read the file
        with NamedTemporaryFile() as tmp:
            workbook.save(tmp.name)
            tmp.seek(0)
            binary_data = tmp.read()
            workbook.close()
            tmp.close()

        return FileResponse(
            io.BytesIO(binary_data),
            as_attachment=True,
            filename="export.xlsx",
        )


class MultilingualDefaultOrdering:
    """Add / modify default ordering for localized fields.

    Even though Django can deal with localized fields and will happily
    add them to "ORDER BY" clause, Postgresql's documentation states that
    this is not "particularly useful". We should sort by a specific language
    instead, even "by default".

    This adds a .order_by() call to the base queryset, using the model's
    default ordering. If the default ordering is a localized field, we
    use the currently used language for it (respecting forced-monolingual
    models as well)

    Each order_by() call will clear any previous ordering, so it's safe
    for us to "force" a default ordering here, it can still be overridden
    in regular ordering filters etc which come after
    """

    def get_queryset(self):
        qs = super().get_queryset()
        lang = utils.forced_or_current_lang(qs.model.__name__)
        ordering = qs.model._meta.ordering

        def _order_field_suffix_if_needed(field_name):
            field = qs.model._meta.get_field(field_name)
            if isinstance(field, localized_fields.fields.LocalizedField):
                return f"{field_name}__{lang}"

            return field_name

        new_ordering = [_order_field_suffix_if_needed(field) for field in ordering]

        qs = qs.order_by(*new_ordering)
        return qs


class ScopeViewSet(MultilingualDefaultOrdering, BaseViewset):
    serializer_class = serializers.ScopeSerializer
    queryset = models.Scope.objects.all()
    search_fields = (
        "name",
        "description",
    )
    multilingual_search_fields = [
        "name",
        "description",
    ]

    case_insensitive_ordering_fields = [
        "name",
        "description",
        "full_name",
    ]
    filterset_class = filters.ScopeFilterset

    def get_queryset(self):
        qs = super().get_queryset()
        qs = self.replace_prefetch(qs, "acls", ACLViewSet)

        return qs


class RoleViewSet(MultilingualDefaultOrdering, BaseViewset):
    serializer_class = serializers.RoleSerializer
    queryset = models.Role.objects.all()
    search_fields = (
        "slug",
        "name",
        "description",
    )
    multilingual_search_fields = [
        "name",
        "description",
    ]

    case_insensitive_ordering_fields = [
        "name",
        "description",
    ]

    def get_queryset(self):
        qs = super().get_queryset()
        return self.replace_prefetch(qs, "acls", ACLViewSet)


class PermissionViewSet(BaseViewset):
    serializer_class = serializers.PermissionSerializer
    queryset = models.Permission.objects.prefetch_related("roles")
    filterset_fields = ("roles",)
    search_fields = (
        "slug",
        "name",
        "description",
    )
    multilingual_search_fields = [
        "name",
        "description",
    ]
    case_insensitive_ordering_fields = [
        "name",
        "description",
    ]


class ACLViewSet(BaseViewset):
    serializer_class = serializers.ACLSerializer
    queryset = models.ACL.objects.all().select_related("user", "scope", "role")
    filterset_class = filters.ACLFilterset
    search_fields = [
        "scope__name",
        "scope__description",
        "role__name",
        "role__description",
        "user__username",
        "user__first_name",
        "user__last_name",
        "=user__zip",
        "user__city",
    ]

    multilingual_search_fields = [
        "scope__name",
        "scope__description",
        "role__name",
        "role__description",
        "user__city",
    ]



# Generated by Django 3.2.12 on 2022-04-11 10:59

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("emeis_core", "0007_change_JSONField_to_native_django_model_field"),
    ]

    operations = [
        migrations.AddField(
            model_name="scope",
            name="is_active",
            field=models.BooleanField(default=True),
        ),
    ]



import os
import re
from warnings import warn

import environ
from django.conf import global_settings
from pkg_resources import resource_filename

env = environ.Env()
django_root = environ.Path(__file__) - 2

ENV_FILE = env.str("ENV_FILE", default=django_root(".env"))
if os.path.exists(ENV_FILE):
    environ.Env.read_env(ENV_FILE)

# per default production is enabled for security reasons
# for development create .env file with ENV=development
ENV = env.str("ENV", "production")


def default(default_dev=env.NOTSET, default_prod=env.NOTSET):
    """Environment aware default."""
    return default_prod if ENV == "production" else default_dev


SECRET_KEY = env.str("SECRET_KEY", default=default("uuuuuuuuuu"))
DEBUG = env.bool("DEBUG", default=default(True, False))
ALLOWED_HOSTS = env.list("ALLOWED_HOSTS", default=default(["*"]))


# Application definition

INSTALLED_APPS = [
    "django.contrib.postgres",
    "localized_fields",
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "mozilla_django_oidc",
    "mptt",
    "generic_permissions.apps.GenericPermissionsConfig",
    "emeis.core.apps.DefaultConfig",
]

if ENV == "dev":
    try:
        __import__("django_extensions")
        INSTALLED_APPS.append("django_extensions")
    except ImportError:  # pragma: no cover
        # Nothing bad, just won't have django-extensions
        # niceties installed (Most likely Emeis was built
        # without dev dependencies)
        pass

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.locale.LocaleMiddleware",
]

ROOT_URLCONF = "emeis.urls"
WSGI_APPLICATION = "emeis.wsgi.application"


# Database
# https://docs.djangoproject.com/en/1.11/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "psqlextra.backend",
        "NAME": env.str("DATABASE_NAME", default="emeis"),
        "USER": env.str("DATABASE_USER", default="emeis"),
        "PASSWORD": env.str("DATABASE_PASSWORD", default=default("emeis")),
        "HOST": env.str("DATABASE_HOST", default="localhost"),
        "PORT": env.str("DATABASE_PORT", default=""),
        "OPTIONS": env.dict("DATABASE_OPTIONS", default={}),
    }
}


# Cache
# https://docs.djangoproject.com/en/1.11/ref/settings/#caches

CACHES = {
    "default": {
        "BACKEND": env.str(
            "CACHE_BACKEND", default="django.core.cache.backends.locmem.LocMemCache"
        ),
        "LOCATION": env.str("CACHE_LOCATION", ""),
    }
}


# Extensions

EMEIS_VISIBILITY_CLASSES = env.list(
    "EMEIS_VISIBILITY_CLASSES", default=(["generic_permissions.visibilities.Any"])
)

EMEIS_PERMISSION_CLASSES = env.list(
    "EMEIS_PERMISSION_CLASSES", default=(["generic_permissions.permissions.AllowAny"])
)

VISIBILITY_CLASSES = env.list(
    "VISIBILITY_CLASSES", default=default(["generic_permissions.visibilities.Any"])
)

PERMISSION_CLASSES = env.list(
    "PERMISSION_CLASSES", default=default(["generic_permissions.permissions.AllowAny"])
)

# Reading PERMISSION_CLASSES, VISIBILITY_CLASSES is still supported. If
# it's set but EMEIS_* is not, copy over the config
if PERMISSION_CLASSES and not EMEIS_PERMISSION_CLASSES:  # pragma: no cover
    EMEIS_PERMISSION_CLASSES = PERMISSION_CLASSES
if VISIBILITY_CLASSES and not EMEIS_VISIBILITY_CLASSES:  # pragma: no cover
    EMEIS_VISIBILITY_CLASSES = VISIBILITY_CLASSES


# VISIBILITY_CLASSES and PERMISSION_CLASSES are deprecated, but
# still supported for now. If they're set, notify the user.
def _deprecate_env(name, replacement):
    if env.str(name, default=False):  # pragma: no cover
        warn(
            DeprecationWarning(
                f"The {name} setting is deprecated and will be removed "
                f"in a future version of Emeis. Use {replacement}"
            )
        )


_deprecate_env("PERMISSION_CLASSES", "EMEIS_PERMISSION_CLASSES")
_deprecate_env("VISIBILITY_CLASSES", "EMEIS_VISIBILITY_CLASSES")


EMEIS_VALIDATION_CLASSES = env.list("EMEIS_VALIDATION_CLASSES", default=[])

# We use DGAP as a permission/visibility/validation handler. Copy
# the configuration over so DGAP knows
GENERIC_PERMISSIONS_VISIBILITY_CLASSES = EMEIS_VISIBILITY_CLASSES
GENERIC_PERMISSIONS_PERMISSION_CLASSES = EMEIS_PERMISSION_CLASSES
GENERIC_PERMISSIONS_VALIDATION_CLASSES = EMEIS_VALIDATION_CLASSES

# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/


def parse_languages(languages):
    return [(language, language) for language in languages]


LANGUAGE_CODE = env.str("LANGUAGE_CODE", "en")
LANGUAGES = (
    parse_languages(env.list("LANGUAGES", default=["en"])) or global_settings.LANGUAGES
)
LOCALE_PATHS = env.list("LOCALE_PATHS", default=[django_root("emeis", "locale")])

_lang_codes = [lang for lang, _ in LANGUAGES]
LOCALIZED_FIELDS_FALLBACKS = {
    lang: [LANGUAGE_CODE] + _lang_codes for lang, _ in LANGUAGES
}
del _lang_codes

# Force models to be monolingual despite them having
# multilingual fields. This causes searches to use a "fixed"
# locale instead of the currently-selected one. Note that
# you MUST set the corresponding setting in ember-emeis as well
# as it will then store the values in a "fixed" locale as well.
#
# If set via env var, set EMEIS_FORCE_MODEL_LOCALE=user:de,scope:de,...
# where the model names must be lowercase
EMEIS_FORCE_MODEL_LOCALE = {
    model: lang
    for path in env.list("EMEIS_FORCE_MODEL_LOCALE", default=[])
    for model, lang in path.split(":")
}


TIME_ZONE = env.str("TIME_ZONE", "UTC")
USE_I18N = True
USE_L10N = True
USE_TZ = True

# Authentication
AUTH_USER_MODEL = "emeis_core.User"

OIDC_OP_USER_ENDPOINT = env.str("OIDC_OP_USER_ENDPOINT", default=None)
OIDC_OP_TOKEN_ENDPOINT = "not supported in emeis, but a value is needed"
OIDC_VERIFY_SSL = env.bool("OIDC_VERIFY_SSL", default=True)
OIDC_USERNAME_CLAIM = env.str("OIDC_USERNAME_CLAIM", default="sub")
OIDC_EMAIL_CLAIM = env.str("OIDC_EMAIL_CLAIM", default="email")
OIDC_BEARER_TOKEN_REVALIDATION_TIME = env.int(
    "OIDC_BEARER_TOKEN_REVALIDATION_TIME", default=0
)
OIDC_CREATE_USER = env.bool("OIDC_CREATE_USER", False)
OIDC_UPDATE_USER = env.bool("OIDC_UPDATE_USER", False)
OIDC_OP_INTROSPECT_ENDPOINT = env.str("OIDC_OP_INTROSPECT_ENDPOINT", default=None)
OIDC_RP_CLIENT_ID = env.str("OIDC_RP_CLIENT_ID", default=None)
OIDC_RP_CLIENT_SECRET = env.str("OIDC_RP_CLIENT_SECRET", default=None)
OIDC_DRF_AUTH_BACKEND = "emeis.oidc_auth.authentication.EmeisAuthenticationBackend"

EMEIS_OIDC_USER_FACTORY = env.str(
    "EMEIS_OIDC_USER_FACTORY", default="emeis.oidc_auth.authentication.OIDCUser"
)

REST_FRAMEWORK = {
    "EXCEPTION_HANDLER": "rest_framework_json_api.exceptions.exception_handler",
    "DEFAULT_PAGINATION_CLASS": "rest_framework_json_api.pagination.JsonApiPageNumberPagination",
    "DEFAULT_PARSER_CLASSES": (
        "rest_framework_json_api.parsers.JSONParser",
        "rest_framework.parsers.JSONParser",
        "rest_framework.parsers.FormParser",
        "rest_framework.parsers.MultiPartParser",
    ),
    "DEFAULT_RENDERER_CLASSES": (
        "rest_framework_json_api.renderers.JSONRenderer",
        "rest_framework.renderers.JSONRenderer",
    ),
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "mozilla_django_oidc.contrib.drf.OIDCAuthentication",
    ),
    "DEFAULT_PERMISSION_CLASSES": (
        "rest_framework.permissions.IsAuthenticatedOrReadOnly",
    ),
    "DEFAULT_METADATA_CLASS": "rest_framework_json_api.metadata.JSONAPIMetadata",
    "DEFAULT_FILTER_BACKENDS": (
        "rest_framework_json_api.filters.QueryParameterValidationFilter",
        "emeis.core.filters.EmeisOrderingFilter",
        "rest_framework_json_api.django_filters.DjangoFilterBackend",
        "emeis.core.filters.EmeisSearchFilter",
    ),
    "SEARCH_PARAM": "filter[search]",
    "ORDERING_PARAM": "sort",
    "TEST_REQUEST_RENDERER_CLASSES": (
        "rest_framework_json_api.renderers.JSONRenderer",
        "rest_framework.renderers.JSONRenderer",
    ),
    "TEST_REQUEST_DEFAULT_FORMAT": "vnd.api+json",
}

JSON_API_FORMAT_FIELD_NAMES = "dasherize"
JSON_API_FORMAT_TYPES = "dasherize"
JSON_API_PLURALIZE_TYPES = True


# Anonymous writing
ALLOW_ANONYMOUS_WRITE = env.bool("ALLOW_ANONYMOUS_WRITE", default=False)

if ALLOW_ANONYMOUS_WRITE:
    REST_FRAMEWORK["DEFAULT_PERMISSION_CLASSES"] = [
        "rest_framework.permissions.AllowAny",
    ]


def parse_admins(admins):
    """
    Parse env admins to django admins.

    Example of ADMINS environment variable:
    Test Example <test@example.com>,Test2 <test2@example.com>
    """
    result = []
    for admin in admins:
        match = re.search(r"(.+) \<(.+@.+)\>", admin)
        if not match:  # pragma: no cover
            raise environ.ImproperlyConfigured(
                'In ADMINS admin "{0}" is not in correct '
                '"Firstname Lastname <email@example.com>"'.format(admin)
            )
        result.append((match.group(1), match.group(2)))
    return result


ADMINS = parse_admins(env.list("ADMINS", default=[]))

USER_EXPORT_TEMPLATE_PATH = env.str(
    "USER_EXPORT_TEMPLATE_PATH",
    default=resource_filename("emeis.core", "templates/user_list.xlsx"),
)

# Cors headers
CORS_ORIGIN_ALLOW_ALL = env.bool("CORS_ORIGIN_ALLOW_ALL", default=False)
CORS_ORIGIN_WHITELIST = env.list("CORS_ORIGIN_WHITELIST", default=[])

# Logging

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {"console": {"class": "logging.StreamHandler"}},
    "loggers": {
        "": {"handlers": ["console"], "level": env.str("LOG_LEVEL", default="INFO")}
    },
}

EMEIS_META_FIELDS = env.list("EMEIS_META_FIELDS", default=[])



import json

import pytest
from django.core.management import call_command
from django.core.management.base import CommandError

from emeis.core.models import ACL, Scope


def _replace_in_text(in_str, **kwargs):
    if isinstance(in_str, dict):
        return {k: _replace_in_text(v, **kwargs) for k, v in in_str.items()}
    for key, val in kwargs.items():
        in_str = in_str.replace(f"${key}", str(val))
    return in_str


def _replace_in_args(cmd_args, user, role, scope):
    if "$role" in cmd_args:
        role_idx = cmd_args.index("$role")
        cmd_args[role_idx] = str(role.pk)
    if "$user" in cmd_args:
        user_idx = cmd_args.index("$user")
        cmd_args[user_idx] = str(user.username)
    if "$scope" in cmd_args:
        scope_idx = cmd_args.index("$scope")
        cmd_args[scope_idx] = str(scope.pk)
    return cmd_args


@pytest.mark.parametrize(
    "cmd_args, expect_out, expect_err, expect_exc",
    [
        ([], "", "", "Error: the following arguments are required: --name/-n"),
        (
            ["--name", "foo", "--user", "foo"],
            "",
            "If you pass --user, you must also pass --role",
            "",
        ),
        (
            ["--name", "foo", "--role", "foo"],
            "",
            "If you pass --role, you must also pass --user",
            "",
        ),
    ],
)
def test_param_validation(
    transactional_db,
    capsys,
    cmd_args,
    expect_out,
    expect_err,
    expect_exc,
    role,
    user,
    scope,
):
    _replace_in_args(cmd_args, user=user, role=role, scope=scope)

    if expect_exc:
        with pytest.raises(CommandError) as exc:
            call_command("create_scope", *cmd_args)

        exc_msg = "".join(exc.value.args)
        assert expect_exc.strip() == exc_msg.strip()

    else:
        call_command("create_scope", *cmd_args)
        stdout, stderr = capsys.readouterr()
        assert stdout.strip() == expect_out.strip()
        assert stderr.strip() == expect_err.strip()


@pytest.mark.parametrize(
    "cmd_args, expect_out, expect_err",
    [
        (["--name", "foobar"], "Scope created: id=$new_scope", ""),
        (
            ["--name", "foobar", "--user", "$user", "--role", "$role"],
            (
                "ACL for new scope created: user=$user, role=$role\n"
                "Scope created: id=$new_scope\n"
            ),
            "",
        ),
        (
            [
                "--name",
                '{"de": "Name DE", "en": "Name EN"}',
                "--user",
                "$user",
                "--role",
                "$role",
            ],
            (
                "ACL for new scope created: user=$user, role=$role\n"
                "Scope created: id=$new_scope\n"
            ),
            "",
        ),
        (
            ["--name", "foobar", "--user", "$user", "--role", "$role", "--json-out"],
            {"scope_id": "$new_scope", "acl_id": "$new_acl"},
            "",
        ),
        (
            [
                "--name",
                "foobar",
                "--user",
                "$user",
                "--role",
                "$role",
                "--parent",
                "$scope",
                "--json-out",
            ],
            {
                "scope_id": "$new_scope",
                "acl_id": "$new_acl",
                "parent_scope_id": "$scope",
            },
            "",
        ),
        (
            [
                "--name",
                "foobar",
                "--user",
                "$user",
                "--role",
                "$role",
                "--parent",
                "f8f635d7-dcc9-4114-838b-b63e2584f686",
                "--json-out",
            ],
            "",
            "Parent scope with given ID not found: f8f635d7-dcc9-4114-838b-b63e2584f686",
        ),
    ],
)
def test_create_success(
    transactional_db, capsys, cmd_args, expect_out, expect_err, role, user, scope, acl
):
    _replace_in_args(cmd_args, user=user, role=role, scope=scope)
    call_command("create_scope", *cmd_args)

    new_scope = Scope.objects.all().order_by("created_at").last()
    new_acl = ACL.objects.all().order_by("created_at").last()

    def replace_expectations(in_str):
        return _replace_in_text(
            in_str,
            scope=scope.pk,
            role=role.pk,
            user=user.username,
            new_scope=new_scope.pk,
            new_acl=new_acl.pk,
        )

    expect_err = replace_expectations(expect_err)
    expect_out = replace_expectations(expect_out)

    if not expect_err:
        # Check that a scope has indeed been created
        assert new_scope != scope

    stdout, stderr = capsys.readouterr()
    if isinstance(expect_out, dict):
        assert json.loads(stdout) == expect_out
    else:
        assert stdout.strip() == expect_out.strip()

    assert stderr.strip() == expect_err.strip()



import mptt
import mptt.managers
from django.conf import settings
from django.utils import translation


def rebuild_mptt_model(model):
    """
    Rebuild an mptt model.

    This is used in migrations that work with data from mptt models.
    """
    manager = mptt.managers.TreeManager()
    manager.model = model
    mptt.register(model)
    manager.contribute_to_class(model, "objects")
    manager.rebuild()


def forced_or_current_lang(model_name):
    """Return the currently to-be-used language for the given model.

    If no monolingual mode is forced for the given model, this will
    return the currently active language as per Django (requested from
    client). However if the model is forced into monolingual mode,
    that language will be returned instead.

    """
    return settings.EMEIS_FORCE_MODEL_LOCALE.get(
        model_name.lower(), translation.get_language()
    )



# Generated by Django 3.2.12 on 2022-04-21 07:24

from django.db import migrations
import django.db.models.deletion
import mptt.fields


class Migration(migrations.Migration):
    dependencies = [
        ("emeis_core", "0008_scope_is_active"),
    ]

    operations = [
        migrations.AlterField(
            model_name="scope",
            name="parent",
            field=mptt.fields.TreeForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="children",
                to="emeis_core.scope",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots["test_user_export 1"] = [
    ["Name", "First Name", "Email", "Roles and scopes"],
    [
        "Marsh",
        "Heather",
        "dmiller@example.org",
        """Barbara Gomez: Cody West
Katherine Gardner: Kevin Wood""",
    ],
    ["Ortega", "Nicole", "vgarcia@example.com", "Michelle Harris: Andrea Cunningham"],
    ["Osborne", "Aaron", "tonymata@example.net", "Christopher Thornton: Brooke Hill"],
    ["Grimes", "Ricky", "david81@example.net", "Zachary Garrett: Andrew Ruiz"],
    ["Morales", "David", "scottvalencia@example.net", "Anne Perkins: Christine Cooper"],
    ["Chan", "Connor", "jacqueline41@example.com", "Robert Barnes: Erica Hawkins"],
    ["Short", "Julia", "vdyer@example.net", "Kevin Franco: Alexis Sanchez"],
    ["Brown", "Robert", "xchase@example.org", "Rebecca Mullins: John Hill"],
    ["Campbell", "Jeremy", "lucas09@example.net", "Alicia Wells: Tracy Wallace MD"],
    [
        "Obrien",
        "Melissa",
        "theresaanderson@example.com",
        "Kristina Davis: David Taylor",
    ],
]



# Generated by Django 2.2.13 on 2020-06-18 07:55

import localized_fields.fields.char_field
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("emeis_core", "0002_data_bootstrap"),
    ]

    operations = [
        migrations.AlterField(
            model_name="user",
            name="city",
            field=localized_fields.fields.char_field.LocalizedCharField(
                blank=True, max_length=255, null=True, required=[], verbose_name="city"
            ),
        ),
    ]
